/*******************************************************************************
 * Copyright (c) 2011 Karlsruhe Institute of Technology (KIT) - Institute for 
 * Applied Computer Science (IAI). 
 * 
 * This file is part of soda4LCA - the Service-Oriented Life Cycle Data Store.
 * 
 * soda4LCA is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the 
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * soda4LCA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with soda4LCA.  If not, see <http://www.gnu.org/licenses/>.
 ******************************************************************************/


package de.iai.ilcd.model.dao;

import java.util.HashMap;
import java.util.List;

import javax.persistence.EntityManager;
import javax.persistence.NoResultException;
import javax.persistence.Query;

import org.apache.velocity.tools.generic.ValueParser;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.fzk.iai.ilcd.service.model.IProcessListVO;
import de.fzk.iai.ilcd.service.model.IProcessVO;
import de.fzk.iai.ilcd.service.model.enums.TypeOfProcessValue;
import de.iai.ilcd.model.common.DataSetVersion;
import de.iai.ilcd.model.common.GeographicalArea;
import de.iai.ilcd.model.common.GlobalReference;
import de.iai.ilcd.model.flow.Flow;
import de.iai.ilcd.model.process.Exchange;
import de.iai.ilcd.model.process.Process;
import de.iai.ilcd.persistence.PersistenceUtil;

/**
 * Data access object for {@link Process processes}
 */
public class ProcessDao extends DataSetDao<Process, IProcessListVO> {

	private static final Logger logger = LoggerFactory.getLogger(ProcessDao.class);

	private final static HashMap<String, String> ORDER_BY_KEY_MAP = ProcessDao.createOrderByKeyMap();
	
	public ProcessDao() {
		super("Process", Process.class, IProcessListVO.class);
	}

	private static HashMap<String, String> createOrderByKeyMap() {
		HashMap<String, String> map = new HashMap<String, String>();
		// values provided by data table lazy model
		map.put("processName.name","p.processName.basePart.defaultValue");
		map.put("type.value","p.type");
		map.put("geography.location","p.geography.location");
		map.put("classification.classHierarchyAsString","c1.name ascDesc,c2.name");
		map.put("timeInformation.referenceYear","p.timeInformation.referenceYear");
		map.put("timeInformation.validUntil","p.timeInformation.validUntil");
		map.put("LCIMethodInformation.methodPrinciple.value","p.lCIMethodInformation.methodPrinciple");
		return map;
	}

	public Process getFullProcess(String uuid) {
		Process process = this.getByUuid(uuid);
		if (process != null) {
			addFlowsToExchanges(process);
		}
		return process;
	}

	@Override
	protected void preCheckAndPersist(Process dataSet) {
		this.addFlowsToExchanges(dataSet);
	}

	@SuppressWarnings("unchecked")
	public List<GeographicalArea> getUsedLocations() {
		EntityManager em = PersistenceUtil.getEntityManager();
		return (List<GeographicalArea>) em
				.createQuery(
						"select distinct area from Process p, GeographicalArea area where p.geography.location=area.areaCode order by area.name")
				.getResultList();
	}

	@SuppressWarnings("unchecked")
	public List<GeographicalArea> getAllLocations() {
		EntityManager em = PersistenceUtil.getEntityManager();
		return (List<GeographicalArea>) em.createQuery(
				"select distinct area from GeographicalArea area order by area.name").getResultList();
	}

	@SuppressWarnings("unchecked")
	public List<Integer> getReferenceYears() {
		EntityManager em = PersistenceUtil.getEntityManager();
		return (List<Integer>) em
				.createQuery(
						"select distinct p.timeInformation.referenceYear from Process p order by p.timeInformation.referenceYear asc")
				.getResultList();
	}

	@Override
	protected Query createQueryObject(ValueParser searchParams, String sortString, boolean ascending, boolean returnCount) {
		// TODO: clean up this mess
		EntityManager em = PersistenceUtil.getEntityManager();

		ValueParser params = new ValueParser();
		if (searchParams != null)
			params = searchParams;

		StringBuffer queryString = new StringBuffer();
		if (returnCount) {
			queryString.append("select count(distinct p) from Process p ");
		} else {
			queryString.append("select distinct p from Process p ");
		}

		// order by string
		String orderSql = ProcessDao.ORDER_BY_KEY_MAP.get( sortString );
		if ( orderSql == null ){
			orderSql = "p.processName.basePart.defaultValue";
		}
		
		// determine if this is a classification order by query (several additional steps are required...)
		boolean classificationOrderBy = "c1.name ascDesc,c2.name".equals(orderSql);
		
		// join is only for order by on classification, not required for count
		if ( !returnCount && classificationOrderBy ){
			orderSql = orderSql.replaceAll("ascDesc", ascending ? "asc" : "desc");
			queryString.append( this.buildClassificationJoin("p") );
		}
		
		queryString.append("where ");
		
		// where statements are only for order by on classification, not required for count
		if ( !returnCount && classificationOrderBy ){
			queryString.append( this.buildClassificationWhere("p") );
			queryString.append(" AND ");
		}
		
		String namePhrase = params.getString("name");
		if (namePhrase == null) {
			namePhrase = "";
		}
		String descriptionPhrase = params.getString("description");
		if (descriptionPhrase == null) {
			descriptionPhrase = "";
		}

		queryString.append("(");

		queryString.append("p.processName.basePart.defaultValue like :namePhrase");
		queryString.append(" or p.processName.routePart.defaultValue like :namePhrase");
		queryString.append(" or p.processName.locationPart.defaultValue like :namePhrase");
		queryString.append(" or p.processName.unitPart.defaultValue like :namePhrase");
		queryString.append(" or p.synonyms.defaultValue like :namePhrase");
		if (descriptionPhrase.length() > 2) {
			queryString.append(" or p.description.defaultValue like :descriptionPhrase");
			queryString.append(" or p.useAdvice.defaultValue like :descriptionPhrase");
		}
		queryString.append(")");

		String[] locations = params.getStrings("location");
		if ((locations != null) && (locations.length > 0)) {
			queryString.append(" and ( ");
			for (int i = 0; i < locations.length; i++) {
				queryString.append("p.geography.location = '").append(locations[i]).append("'");
				if (i < locations.length - 1) {
					queryString.append(" or ");
				}
			}
			queryString.append(" )");
		}

		// Get the integer values (no or illegal value => null)
		Integer lowerRefYear = this.parseInteger(params.getString("referenceYearLower"));
		Integer upperRefYear = this.parseInteger(params.getString("referenceYearUpper"));
		Integer lowerValidUntil = this.parseInteger(params.getString("lowerValidUntil"));
		Integer upperValidUntil = this.parseInteger(params.getString("upperValidUntil"));

		if (lowerRefYear != null) {
			queryString.append(" and ( p.timeInformation.referenceYear >= ").append(lowerRefYear).append(")");
		}

		if (upperRefYear != null) {
			queryString.append(" and ( p.timeInformation.referenceYear <= ").append(upperRefYear).append(")");
		}

		if (lowerValidUntil != null) {
			queryString.append(" and ( p.timeInformation.validUntil >= ").append(lowerValidUntil).append(")");
		}

		if (upperValidUntil != null) {
			queryString.append(" and ( p.timeInformation.validUntil <= ").append(upperValidUntil).append(")");
		}

		String type = params.getString("type");
		if (type != null && (type.length() > 3) && (!type.equals("select option"))) {
			TypeOfProcessValue typeValue = null;
			try {
				typeValue = TypeOfProcessValue.fromValue(type);
			} catch (Exception e) {
				// ignore it as we do not have a parsable value
			}
			if (typeValue != null)
				queryString.append(" and ( p.type like '").append(typeValue.toString()).append("')");
		}

		if (params.getString("parameterized") != null) {
			queryString.append(" and p.parameterized=true");
		}

		// now let's try to handle categories
		String[] categories = params.getStrings("classes");
		// logger.info("categories contains {} categories",categories.length);

		if ((categories != null) && (categories.length > 0)) {
			if (!(categories.length == 1 && categories[0].equals("all"))) {
				queryString.append(" and ( ");
				for (int i = 0; i < categories.length; i++) {
					if (!categories[i].equals("all")) {
						queryString.append("('" + categories[i] + "'" + " in (select c" + i
								+ ".name from p.classification.classes c" + i + "))");
						if (i < categories.length - 1) {
							queryString.append(" or ");
						}
					}
				}
				queryString.append(" )");
			}
		}

		/*
		 * String methodPrinciple=params.getString("methodPrinciple"); if (methodPrinciple != null &&
		 * (methodPrinciple.length() > 3)) { queryString.append(" and ( dataset.typeOfProcess like " + methodPrinciple +
		 * ")"); }
		 * 
		 * String completeness=params.getString("completeness"); if (completeness != null && (completeness.length() >
		 * 3)) { queryString.append(" and ( dataset.dsDetails.completeness like " + completeness + ")"); }
		 */


		// order by only on non-count queries
		if ( !returnCount ){
			queryString.append(" order by " + orderSql + " " + (ascending ? "asc":"desc") );
		}
		
		logger.debug("search query: {}", queryString);
//		if (sortString != null)
//			logger.error("sortString = " + sortString);

		Query query;

		if (descriptionPhrase.length() > 2) {
			query = em.createQuery(queryString.toString()).setParameter("namePhrase", "%" + namePhrase + "%")
					.setParameter("descriptionPhrase", "%" + descriptionPhrase + "%");
		} else {
			query = em.createQuery(queryString.toString()).setParameter("namePhrase", "%" + namePhrase + "%");
		}

		return query;
	}

	private Integer parseInteger(String s) {
		if (s == null || s.trim().isEmpty()) {
			return null;
		}
		try {
			return new Integer(s);
		} catch (Exception e) {
			return null;
		}
	}

	public IProcessVO getForeignProcess(String nodeShortName, String uuid) {
		ForeignDataSetsHelper foreignHelper = new ForeignDataSetsHelper();
		return foreignHelper.getForeignDataSet(IProcessVO.class, nodeShortName, uuid);
	}

	private void addFlowsToExchanges(Process process) {
		FlowDao flowDao = new FlowDao();

		for (Exchange exchange : process.getExchanges()) {

			if (exchange.getFlow() != null) {
				continue; // flow already associated
			}
			if (exchange.getFlowReference() == null) {
				continue;
			}
			GlobalReference flowReference = exchange.getFlowReference();
			String flowUuid = null;
			if (flowReference.getUuid() == null) {
				// crap, no uuid attribute in flow reference
				// but we can try to get it from the uri attribute if there
				String uri = flowReference.getUri();
				if (uri == null) {
					continue; // sorry we can't get uuid
				}
				String[] splittedUri = uri.split("_");
				if (splittedUri.length <= 2) {
					continue; // uri does not contain enough parts: so no uuid
				}
				flowUuid = splittedUri[splittedUri.length - 2];
			} else {
				flowUuid = exchange.getFlowReference().getUuid().getUuid();
			}
			Flow flow = null;
			try {
				flow = flowDao.getByUuid(flowUuid);
				exchange.setFlow(flow);
			} catch (NoResultException ex) {
				// OK, we have no associated flow, so do nothing
			}
		}
	}

	/*
	 * ================================================== == DEPRECATED METHODS / BACKWARDS COMPATIBILITY ==
	 * ==================================================
	 */

	/**
	 * Get list of all processes
	 * 
	 * @return list of all processes
	 * @deprecated still exists for internal backwards compatibility, may be removed in future releases. Use
	 *             {@link #getAll()}.
	 * @see #getAll()
	 */
	@Deprecated
	public List<Process> getProcesses() {
		return this.getAll();
	}

	/**
	 * Get a process by UUID
	 * 
	 * @param uuid
	 *            UUID of flow to get
	 * @return process with provided UUID
	 * @deprecated still exists for internal backwards compatibility, may be removed in future releases. Use
	 *             {@link #getByUuid(String)}.
	 * @see #getByUuid(String)
	 */
	@Deprecated
	public Process getProcess(String uuid) {
		return this.getByUuid(uuid);
	}

	/**
	 * Get {@link Process} by UUID string and version
	 * 
	 * @param uuid
	 *            the UUID string
	 * @param version
	 *            the version of the method
	 * @return {@link Process} for the specified UUID string/version, or <code>null</code> if none found
	 * @deprecated still exists for internal backwards compatibility, may be removed in future releases. Use
	 *             {@link #getByUuidAndVersion(String, DataSetVersion)}.
	 */
	@Deprecated
	public Process getProcess(String uuid, DataSetVersion version) {
		return this.getByUuidAndVersion(uuid, version);
	}

	/**
	 * Get the process with the provided JPA id
	 * 
	 * @param datasetId
	 *            JPA id
	 * @return process with the provided JPA id
	 * @deprecated still exists for internal backwards compatibility, may be removed in future releases. Use
	 *             {@link #getByDataSetId(long)}.
	 * @see #getByDataSetId(long)
	 */
	@Deprecated
	public Process getProcess(long datasetId) {
		return this.getByDataSetId(datasetId);
	}

	/**
	 * Get the process with the provided JPA id
	 * 
	 * @param id
	 *            JPA id as string
	 * @return process with the provided JPA id
	 * @deprecated still exists for internal backwards compatibility, may be removed in future releases. Use
	 *             {@link #getByDataSetId(String)}.
	 * @see #getByDataSetId(String)
	 */
	@Deprecated
	public Process getProcessById(String id) {
		return this.getByDataSetId(id);
	}

	/**
	 * Get the process with the provided JPA id
	 * 
	 * @param id
	 *            JPA id
	 * @return process with the provided JPA id
	 * @deprecated still exists for internal backwards compatibility, may be removed in future releases. Use
	 *             {@link #getByDataSetId(long)}.
	 * @see #getByDataSetId(long)
	 */
	@Deprecated
	public Process getProcessById(Long id) {
		return this.getByDataSetId(id);
	}

	/**
	 * Get processes by main class
	 * 
	 * @param mainClass
	 *            main class to get process by
	 * @return matched data sets
	 * @deprecated still exists for internal backwards compatibility, may be removed in future releases. Use
	 *             {@link #getByClass(String)}.
	 * @see #getByClass(String)
	 */
	@Deprecated
	public List<Process> getProcessesByClass(String mainClass) {
		return this.getByClass(mainClass);
	}

	/**
	 * Get the number of processes by main class
	 * 
	 * @param mainClass
	 *            main class to get process by
	 * @return the number of matched data sets
	 * @deprecated still exists for internal backwards compatibility, may be removed in future releases. Use
	 *             {@link #getByClass(String)}.
	 * @see #getNumberByClass(String)
	 */
	@Deprecated
	public long getNumberOfProcesses(String mainClass) {
		return this.getNumberByClass(mainClass);
	}

	/**
	 * Get processes by main and sub class
	 * 
	 * @param mainClass
	 *            main class to get process by
	 * @param subClass
	 *            sub class to get process by
	 * @return matched data sets
	 * @deprecated still exists for internal backwards compatibility, may be removed in future releases. Use
	 *             {@link #getByClass(String,String)}.
	 * @see #getByClass(String,String)
	 */
	@Deprecated
	public List<Process> getProcessesByClass(String mainClass, String subClass) {
		return this.getByClass(mainClass, subClass);
	}

	/**
	 * Get number of processes by main and sub class
	 * 
	 * @param mainClass
	 *            main class to get process by
	 * @param subClass
	 *            sub class to get process by
	 * @return number of matched data sets
	 * @deprecated still exists for internal backwards compatibility, may be removed in future releases. Use
	 *             {@link #getByClass(String,String)}.
	 * @see #getNumberByClass(String,String)
	 */
	@Deprecated
	public long getNumberOfProcesses(String mainClass, String subClass) {
		return this.getNumberByClass(mainClass, subClass);
	}

}
