/*******************************************************************************
 * Copyright (c) 2011 Karlsruhe Institute of Technology (KIT) - Institute for 
 * Applied Computer Science (IAI). 
 * 
 * This file is part of soda4LCA - the Service-Oriented Life Cycle Data Store.
 * 
 * soda4LCA is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the 
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * soda4LCA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with soda4LCA.  If not, see <http://www.gnu.org/licenses/>.
 ******************************************************************************/


package de.iai.ilcd.webgui.controller;

import java.net.InetAddress;
import java.net.URI;

import javax.faces.bean.ApplicationScoped;
import javax.faces.bean.ManagedBean;
import javax.faces.context.FacesContext;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.iai.ilcd.configuration.ConfigurationService;
import javax.faces.context.ExternalContext;
import javax.servlet.http.HttpServletRequest;
import org.apache.commons.configuration.Configuration;

/**
 * Bean to access configuration, is in application scope, all values loaded only once
 */
@ManagedBean(name = "conf")
@ApplicationScoped
public class ConfigurationBean {

	private static Logger logger = LoggerFactory.getLogger(ConfigurationBean.class);

        private static Configuration conf=ConfigurationService.getProperties();
        
	/**
	 * Path to templates
	 */
	private String templatePath;

	/**
	 * Name of directory with jQuery UI theme
	 */
	private String themeDirectory;

	/**
	 * Context path
	 */
	private String contextPath;
        
        private int requestPort;

	/**
	 * Base URI
	 */
	private URI baseUri;

	/**
	 * Flag if registration is activated
	 */
	private boolean registrationActivated;
        
        /**
         * Flag if the registration settings allow the user himself to activate his new user acccount
         */
        private boolean selfActivation;

	/**
	 * Path to templates
	 */
	private String applicationTitle;

	/**
	 * Path to templates
	 */
	private String logoPath;
	
	/**
	 * Initialize configuration bean
	 */
	public ConfigurationBean() {
		initializeRequestContextByFacelets();
                initializeOtherFields();
	}
        
        /*
         * Use this constructor if not in the context of JSF; i.e. initialize context path by yourself
         */
        public ConfigurationBean(HttpServletRequest request) {
            this.contextPath=request.getContextPath();
            this.requestPort=request.getServerPort();
            initializeOtherFields();
        }
        
        private void initializeOtherFields() {
		this.templatePath = buildTemplatePath();
		this.themeDirectory = conf.getString("theme");
		this.applicationTitle = conf.getString("title");
		
		this.baseUri = buildBaseUri();
		this.registrationActivated = conf.getBoolean("user.registration.activated",
				false);
                this.selfActivation=conf.getBoolean("user.registration.selfActivation", false);
		
		String logoPath = conf.getString("logo");
		if ( logoPath == null || logoPath.toLowerCase().trim().equals("false") ){
			logoPath = null;
		} else {
			logoPath = logoPath.replace("%contextPath%", this.contextPath);
		}
		this.logoPath = logoPath;            
        }

	/**
	 * Build the context path
	 * 
	 * @return context path
	 */
	private void initializeRequestContextByFacelets() {
                ExternalContext extContext=FacesContext.getCurrentInstance().getExternalContext();
		contextPath=extContext.getRequestContextPath();
                requestPort = extContext.getRequestServerPort();

	}
        

	/**
	 * Build up the template path
	 * @return created template path
	 */
	private String buildTemplatePath() {
		String themeName = conf.getString("template");
		StringBuilder pathBuilder = new StringBuilder();
		pathBuilder.append("/templates/").append(themeName);
		return pathBuilder.toString();
	}

	/**
	 * Build up the base URI
	 * 
	 * @param contextPath
	 *            context path of the application
	 * @return build base URI
	 */
	private URI buildBaseUri() {
		String hostName = null;

		try {
			InetAddress inetAddress = InetAddress.getLocalHost();
			hostName = inetAddress.getHostName();
		} catch (Exception e) {
			logger.error("Cannot find official hostname of local computer", e);
		}

		// if we have an other hostname configured in conf file, let's use that (proxy) name
		String confHostName = conf.getString("service.url.hostname");
		int confPort = conf.getInt("service.url.port", requestPort);
		if (confHostName != null) {
			hostName = confHostName;
			requestPort = confPort;
		}
		if (hostName == null) {
			logger.error("Cannot find any hostname for this computer for URI generation");
			return null;
		}
		// OK, now let's create the URI
		URI uri = null;
		try {
			uri = new URI("http", null, hostName, requestPort, contextPath, null, null);
		} catch (Exception e) {
			logger.error("Cannot generate baseURI for the application", e);
		}
		return uri;
	}        

	/**
	 * Get the template path
	 * 
	 * @return template path
	 */
	public String getTemplatePath() {
		return this.templatePath;
	}

	/**
	 * Get the jQuery UI Theme directory name
	 * 
	 * @return jQuery UI Theme directory name
	 */
	public String getThemeDirectory() {
		return this.themeDirectory;
	}

	/**
	 * Get the context path
	 * 
	 * @return context path
	 */
	public String getContextPath() {
		return this.contextPath;
	}

	/**
	 * Get the base URI
	 * 
	 * @return base URI
	 */
	public URI getBaseUri() {
		return this.baseUri;
	}

	/**
	 * Get the flag if registration is activated
	 * 
	 * @return <code>true</code> if registration shall be allowed, else <code>false</code>
	 */
	public boolean isRegistrationActivated() {
		return this.registrationActivated;
	}

	/**
	 * Convenience method for JSF, delegates to {@link #isRegistrationActivated()}
	 * 
	 * @return {@link #isRegistrationActivated()}
	 */
	public boolean getRegistrationActivated() {
		return this.isRegistrationActivated();
	}
        
        /**
         * Get the flag if the user is allowed to activate himself a newly created user account
         * @return <code>true</code> if self activation is allowed, else <code>false</code>
         */
        public boolean isSelfActivation() {
            return this.selfActivation;
        }
        
        /**
         * Convenience method for JSF, delegates to {@link #isSelfActivation()}
         * @return <code>true</code> if self activation is allowed, else <code>false</code>
         */
        public boolean getSelfActivation() {
            return this.isSelfActivation();
        }
	
	/**
	 * Determine, if logo was provided (<code>{@link #getLogoPath()} != null</code>)
	 * @return <code>true</code> if logo provided, else <code>false</code>
	 */
	public boolean isLogoProvided(){
		return this.logoPath != null;
	}
	
	/**
	 * Get the logo path
	 * @return logo path
	 */
	public String getLogoPath() {
		return this.logoPath;
	}
	
	/**
	 * Get the title of the application
	 * @return title of the application
	 */
	public String getApplicationTitle() {
		return this.applicationTitle;
	}
}
