/*******************************************************************************
 * Copyright (c) 2011 Karlsruhe Institute of Technology (KIT) - Institute for 
 * Applied Computer Science (IAI). 
 * 
 * This file is part of soda4LCA - the Service-Oriented Life Cycle Data Store.
 * 
 * soda4LCA is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the 
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * soda4LCA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with soda4LCA.  If not, see <http://www.gnu.org/licenses/>.
 ******************************************************************************/
package de.iai.ilcd.webgui.controller.admin;

import java.io.Serializable;
import java.util.List;

import javax.faces.application.FacesMessage;
import javax.faces.bean.ManagedBean;
import javax.faces.bean.ViewScoped;
import javax.faces.context.FacesContext;
import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.Transport;

import org.apache.commons.configuration.Configuration;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.iai.ilcd.configuration.ConfigurationService;
import de.iai.ilcd.model.common.GeographicalArea;
import de.iai.ilcd.model.dao.GeographicalAreaDao;
import de.iai.ilcd.model.dao.UserDao;
import de.iai.ilcd.model.security.Address;
import de.iai.ilcd.model.security.IndustrialSector;
import de.iai.ilcd.model.security.User;
import de.iai.ilcd.security.IlcdSecurityRealm;
import de.iai.ilcd.security.UserAccessBean;

@ManagedBean
@ViewScoped
public class UserHandler implements Serializable {

	/**
	 * Serialization ID
	 */
	private static final long serialVersionUID = 8636585943955487402L;

	/**
	 * Logger
	 */
	private static final Logger logger = LoggerFactory.getLogger(UserDao.class);

	/**
	 * Current user instance
	 */
	private User user = new User();

	/**
	 * Verify password value
	 */
	private String verifyPassword = "";

	/**
	 * Plain password
	 */
	private String plainPassword = null;

	/**
	 * Value of spam slider
	 */
	private int spamSliderValue = 0;

	/**
	 * Chosen Sector
	 */
	private int chosenSector = 0;

	/**
	 * List of sectors
	 */
	private List<IndustrialSector> industrialSectors = null;

	/**
	 * List of countries
	 */
	private List<GeographicalArea> countries = null;

	/**
	 * Chosen country
	 */
	private int chosenCountry = 0;
	
	/**
	 * DAO for users
	 */
	private final UserDao dao = new UserDao();

	/** Creates a new instance of UserHandler */
	public UserHandler() {
		// initialize industrialSectors list
		this.industrialSectors = this.dao.getIndustrialSectors();
		GeographicalAreaDao gDao = new GeographicalAreaDao();
		this.countries = gDao.getCountries();

		// check if the page was called with a request parameter specifying the current user
		String userId = FacesContext.getCurrentInstance().getExternalContext().getRequestParameterMap().get("userId");
		if (userId != null) {
			User existingUser = this.dao.getUser(Long.parseLong(userId));
			if (existingUser != null) {
				user = existingUser;
				logger.info("found user with username {} and id {}", user.getUserName(), user.getId());
			}
		}
		
		// logger.error("View info of view is: {}", FacesContext.getCurrentInstance().getViewRoot().getViewId());
		String viewId = FacesContext.getCurrentInstance().getViewRoot().getViewId();
		if (viewId.contains("profile.xhtml")) {
			// OK, we are loaded by a profile page
			// first check if there is a valid user
			UserAccessBean callingUser = new UserAccessBean();
			if (!callingUser.isLoggedIn()) {
				FacesMessage facesMsg = new FacesMessage(FacesMessage.SEVERITY_WARN,
						"A user can change his profile only, if he is loggend in", "");
				FacesContext.getCurrentInstance().addMessage("errorInfo", facesMsg);
				return;
			}
			User currentUser = this.dao.getUser(callingUser.getUserName());
			if (currentUser == null) {
				FacesMessage facesMsg = new FacesMessage(FacesMessage.SEVERITY_WARN, "The user with name "
						+ callingUser.getUserName() + " has no associated profile in the database", "");
				FacesContext.getCurrentInstance().addMessage("errorInfo", facesMsg);
				return;
			}
			this.user = currentUser;

		}
		// don't forget to save password, in case password field is empty
//		this.dbPasswordHash = this.user.getPasswordHash();
		// also, don't forget to set index of chosen sector and country
		IndustrialSector sector = user.getSector();
		if (sector != null) {
			chosenSector = sector.getId().intValue();
		}
		if (this.user.getAddress() != null) {
			String countryCode = user.getAddress().getCountry();
			if (countryCode != null) {
				GeographicalArea country = gDao.getArea(countryCode);
				if (country != null) {
					chosenCountry = country.getId().intValue();
				}
			}
		} else {
			this.user.setAddress(new Address());
		}

	}

	public User getUser() {
		return this.user;
	}

	public void setUser(User user) {
		this.user = user;
	}

	public String getVerifyPassword() {
		return this.verifyPassword;
	}

	public void setVerifyPassword(String verifyPassword) {
		this.verifyPassword = verifyPassword;
	}

	public int getSpamSliderValue() {
		return this.spamSliderValue;
	}

	public void setSpamSliderValue(int spamSliderValue) {
		this.spamSliderValue = spamSliderValue;
	}

	public int getChosenSector() {
		return this.chosenSector;
	}

	public void setChosenSector(int chosenSector) {
		this.chosenSector = chosenSector;
	}

	public List<IndustrialSector> getIndustrialSectors() {
		return this.industrialSectors;
	}

	public int getChosenCountry() {
		return this.chosenCountry;
	}

	public void setChosenCountry(int chosenCountry) {
		this.chosenCountry = chosenCountry;
	}

	public List<GeographicalArea> getCountries() {
		return this.countries;
	}
	
	public String getPlainPassword() {
		return this.plainPassword;
	}
	
	public void setPlainPassword(String plainPassword) {
		this.plainPassword = plainPassword;
	}

	public void createUser() {
		UserAccessBean creatingUser = new UserAccessBean();
		if (!(creatingUser.hasRole("MANAGE_USER") || creatingUser.isSuperAdmin())) {
			FacesMessage facesMsg = new FacesMessage(FacesMessage.SEVERITY_WARN,
					"Only users with rights to manage user accounts can create new users", "");
			FacesContext.getCurrentInstance().addMessage("errorInfo", facesMsg);
			return;
		}

		if (this.user.getUserName() == null) {
			FacesMessage facesMsg = new FacesMessage(FacesMessage.SEVERITY_WARN, "There must be a unique user name!",
					"");
			FacesContext.getCurrentInstance().addMessage("errorInfo", facesMsg);
		}
		if (this.plainPassword == null || this.plainPassword.trim().isEmpty() ) {
			FacesMessage facesMsg = new FacesMessage(FacesMessage.SEVERITY_WARN, "You need to supply a password", "");
			FacesContext.getCurrentInstance().addMessage("errorInfo", facesMsg);
		}
		if (this.verifyPassword == null || !this.verifyPassword.equals(this.plainPassword)) {
			FacesMessage facesMsg = new FacesMessage(FacesMessage.SEVERITY_WARN,
					"password and repeated password are not the same", "");
			FacesContext.getCurrentInstance().addMessage("errorInfo", facesMsg);
		}
		User existingUser = this.dao.getUser(this.user.getUserName());
		if (existingUser != null) {
			FacesMessage facesMsg = new FacesMessage(FacesMessage.SEVERITY_WARN,
					"A user with this user name already exists in the system", "");
			FacesContext.getCurrentInstance().addMessage("errorInfo", facesMsg);
		}
		
		// let's generate a new salt
		this.user.setPasswordHashSalt( this.generateSalt() );
		// let's encrypt the password
		this.user.setPasswordHash( IlcdSecurityRealm.getEncryptedPassword( this.plainPassword, this.user.getPasswordHashSalt()) );

		// set sector and country information in user object
		this.setSectorAndCountry();

		try {
			this.dao.persist(this.user);
			FacesMessage facesMsg = new FacesMessage(FacesMessage.SEVERITY_INFO, "The user account has been created",
					"");
			FacesContext.getCurrentInstance().addMessage("info", facesMsg);
		} catch (Exception e) {
			FacesMessage facesMsg = new FacesMessage(FacesMessage.SEVERITY_WARN,
					"There were some erros while creating the user account; please try again", "");
			FacesContext.getCurrentInstance().addMessage("errorInfo", facesMsg);
		}

	}

	public String newUserRegistration() {
		Configuration conf = ConfigurationService.getProperties();

		// check properties first
		if ( !conf.getBoolean("user.registration.activated", Boolean.FALSE) ){
			FacesMessage facesMsg = new FacesMessage(FacesMessage.SEVERITY_ERROR, "Registration is not permitted!",
			"");
			FacesContext.getCurrentInstance().addMessage("errorInfo", facesMsg);
			return null;
		}
		
		UserDao dao = new UserDao();

		if (this.user.getUserName() == null) {
			FacesMessage facesMsg = new FacesMessage(FacesMessage.SEVERITY_WARN, "There must be a unique user name!",
					"");
			FacesContext.getCurrentInstance().addMessage("errorInfo", facesMsg);
			return null;
		}

		User existingUser = dao.getUser(this.user.getUserName());
		if (existingUser != null) {
			FacesMessage facesMsg = new FacesMessage(FacesMessage.SEVERITY_WARN,
					"A user with this user name already exists in the system; choose another user name", "");
			FacesContext.getCurrentInstance().addMessage("errorInfo", facesMsg);
			return null;
		}

		// set sector in country information in user object
		setSectorAndCountry();

		// generate password, salt and encrypted password and activation key
		String plainPassword = RandomPassword.getPassword(8);
		this.user.setPasswordHashSalt( this.generateSalt() );
		this.user.setPasswordHash( IlcdSecurityRealm.getEncryptedPassword(plainPassword, this.user.getPasswordHashSalt()) );
		
		String registrationKey = RandomPassword.getPassword(20);
		this.user.setRegistrationKey(registrationKey);

		// generate Email messages
		Message userMessage = null;
		Message providerMessage = null;
		RegistrationEmailGenerator emailGenerator = new RegistrationEmailGenerator();
		boolean userSelfActivation = false;
		// get property, how user activation should be performed
		userSelfActivation = conf.getBoolean("user.registration.selfActivation", Boolean.FALSE);
		try {
			if (userSelfActivation) {
				// we send only an email to the user because he can activate his account himself
				userMessage = emailGenerator.genActivationEmailForUser(this.user, plainPassword);
			} else {
				// user only get's a notification about the completed registration
				userMessage = emailGenerator.genUserNotice(this.user,plainPassword);
				// provider get's an email to activate the user
				providerMessage = emailGenerator.genActivationMailForProvider(this.user);
			}
		} catch (MessagingException ex) {
			// OK, cannot generate email, let's show the error message to the user
			FacesMessage facesMsg = new FacesMessage(FacesMessage.SEVERITY_WARN, ex.getMessage(), "");
			FacesContext.getCurrentInstance().addMessage("errorInfo", facesMsg);
			logger.error(ex.getMessage(), ex);
			return null;
		}

		try {
			dao.persist(this.user);
			if (userMessage != null) {
				Transport.send(userMessage);
			}
			if (providerMessage != null) {
				Transport.send(providerMessage);
			}
		} catch (Exception ex) {
			dao.remove(this.user);
			FacesMessage facesMsg = new FacesMessage(FacesMessage.SEVERITY_WARN,
					"There was an unknown failure in the user registration process; please try again", "");
			FacesContext.getCurrentInstance().addMessage("errorInfo", facesMsg);
			logger.error("There was an error while saving registration information or sending registration email", ex);
			return null;
		}

		return "registrated";

	}

	private void setSectorAndCountry() {
		// now, set sector and country
		IndustrialSector sector = null;
		for (IndustrialSector s : industrialSectors) {
			if (s.getId().intValue() == chosenSector) {
				sector = s;
			}
		}
		if (sector != null) {
			this.user.setSector(sector);
		}
		GeographicalArea country = null;
		for (GeographicalArea c : countries) {
			if (c.getId().intValue() == chosenCountry) {
				country = c;
			}
		}
		if (country != null) {
			this.user.getAddress().setCountry(country.getAreaCode());
		}
	}

	public void changeUser() {

		UserAccessBean changingUser = new UserAccessBean();
		if (changingUser == null || !changingUser.isLoggedIn()) {
			FacesMessage facesMsg = new FacesMessage(
					FacesMessage.SEVERITY_WARN,
					"Error: No user logged in! Only the user itself or a user with user management rights can change user entries",
					"");
			FacesContext.getCurrentInstance().addMessage("errorInfo", facesMsg);
			return;
		}
		if (!(changingUser.isSuperAdmin() || changingUser.hasRole("MANAGE_USER") || changingUser.getUserName().equals(
				this.user.getUserName()))) {
			FacesMessage facesMsg = new FacesMessage(FacesMessage.SEVERITY_WARN,
					"Only the user itself or a user with user management rights can change user entries", "");
			FacesContext.getCurrentInstance().addMessage("errorInfo", facesMsg);
			return;
		}

		if (this.user.getUserName() == null) {
			FacesMessage facesMsg = new FacesMessage(FacesMessage.SEVERITY_WARN, "There must be a unique user name!",
					"");
			FacesContext.getCurrentInstance().addMessage("errorInfo", facesMsg);
			return;
		}
		if (this.plainPassword == null || this.plainPassword.trim().isEmpty()) {
			// OK, we don't change the password
		}
		// check if verify entry and plain text password are the same
		else if (this.verifyPassword == null || !this.verifyPassword.equals( this.plainPassword ) ) {
			FacesMessage facesMsg = new FacesMessage(FacesMessage.SEVERITY_WARN,
					"password and repeated password are not the same", "");
			FacesContext.getCurrentInstance().addMessage("errorInfo", facesMsg);
			return;
		}
		// OK, let's change the PW
		else {
			// new salt
			this.user.setPasswordHashSalt( this.generateSalt() );
			// encrpyt new pw
			this.user.setPasswordHash( IlcdSecurityRealm.getEncryptedPassword(this.plainPassword, this.user.getPasswordHashSalt()) );
		}

		// set sector and country information in user object
		this.setSectorAndCountry();

		User existingUser = this.dao.getUser(this.user.getUserName());
		if (existingUser == null) {
			FacesMessage facesMsg = new FacesMessage(FacesMessage.SEVERITY_WARN, "A user with the user name "
					+ this.user.getUserName() + " does not exist", "");
			FacesContext.getCurrentInstance().addMessage("errorInfo", facesMsg);
			return;
		}

		logger.info("Merge on user {} with user id {}", this.user.getUserName(), this.user.getId());

		try {
			this.dao.merge(this.user);
			FacesMessage facesMsg = new FacesMessage(FacesMessage.SEVERITY_INFO, "The user account has been changed",
					"");
			FacesContext.getCurrentInstance().addMessage("info", facesMsg);
		} catch (Exception e) {
			FacesMessage facesMsg = new FacesMessage(FacesMessage.SEVERITY_WARN,
					"There was an error while saving the data; please try again", "");
			FacesContext.getCurrentInstance().addMessage("errorInfo", facesMsg);
		}

	}
	
	/**
	 * Generate a new salt (ensures correct length)
	 * @return new salt
	 */
	private String generateSalt(){
		return RandomPassword.getPassword(20);
	}
}
